#ifndef __GSK_LISP_NAMESPACE_H_
#define __GSK_LISP_NAMESPACE_H_

typedef struct _GskLispNamespace GskLispNamespace;
typedef struct _GskLispMarshal GskLispMarshal;
typedef struct _GskLispCodec GskLispCodec;

typedef void (*GskLispFunc) ();
#define GSK_LISP_FUNC(func)		((GskLispFunc)(func))


#include "gsklispnode.h"
#include "gsklispinput.h"
#include "../gskgtk.h"

typedef gpointer (*GskLispCompiler) (gpointer          compiler_data,
				     GtkType           type,
				     GskLispNode      *node,
				     GskLispNamespace *ns,
				     char            **err_msg);
typedef gboolean (*GskLispRunner)   (GskLispFunc       func,
				     gpointer          func_data,
				     GtkArg           *result,
				     GskLispNamespace *ns,
				     gpointer          compiled_data,
				     char            **err_msg);

struct _GskLispMarshal
{
  GskLispCompiler compiler;
  GskLispRunner runner;
  GDestroyNotify compiled_data_destroy;
};


GskLispNamespace *gsk_lisp_namespace_new       ();
void              gsk_lisp_namespace_add_sub   (GskLispNamespace *ns,
						const char       *name,
						GskLispNamespace *subns);
void              gsk_lisp_namespace_add_value (GskLispNamespace *ns,
						GtkArg           *arg);
gboolean          gsk_lisp_namespace_lookup_val(GskLispNamespace *ns,
						guint             num_names,
						const char       *names[],
						GtkArg           *result);
void gsk_lisp_namespace_ref (GskLispNamespace *ns);
void gsk_lisp_namespace_unref (GskLispNamespace *ns);


/* --- defining new functions --- */
void              gsk_lisp_namespace_add_func  (GskLispNamespace *ns,
						const char       *name,
						GskLispFunc       func,
						GskLispMarshal   *marshal,
						gpointer          compiler_data,
						GDestroyNotify    compiler_data_destroy,
						gpointer          func_data,
						GDestroyNotify    destroy);
void              gsk_lisp_namespace_add_func_c(GskLispNamespace *ns,
						const char       *name,
						GskLispFunc       func,
						GskLispMarshal   *marshal);
typedef struct _GskLispArgSpec GskLispArgSpec;
struct _GskLispArgSpec
{
  guint             num_args;
  GtkArg           *args;		/* the values are ignored */

  /* functions with last_arg_repeats == TRUE can only be implemented
   * in C */
  gboolean          last_arg_repeats;
};
GskLispArgSpec *gsk_lisp_arg_spec_make      (guint           num_args,
					     ...);           
GskLispArgSpec *gsk_lisp_arg_spec_make_perm (guint           num_args, ...);
void            gsk_lisp_arg_spec_free      (GskLispArgSpec *arg_spec);

void  gsk_lisp_namespace_add_func_by_arg_spec (GskLispNamespace *ns,
					       const char       *func_name,
					       GtkType           return_type,
					       GskLispArgSpec   *arg_spec,
					       GskLispFunc       func,
					       gpointer          func_data,
					       GDestroyNotify    destroy);

struct _GskLispCodec
{
  /*< private >*/
  GskLispRunner     runner;
  gpointer          compiled_data;
  GDestroyNotify    compiled_data_destroy;
  GskLispFunc       func;
  gpointer          func_data;
  GtkType           return_type;
};

GskLispCodec     *gsk_lisp_namespace_compile   (GskLispNamespace *ns,
						GskLispNode      *node,
						GtkType           type,
						char            **err_msg);
gboolean          gsk_lisp_codec_run           (GskLispCodec     *codec,
		                                GtkArg           *result,
		                                GskLispNamespace *ns,
						char            **err_msg);
void              gsk_lisp_codec_destroy       (GskLispCodec     *codec);

void              gsk_lisp_namespace_nick      (GskLispNamespace *ns,
						const char       *nickname,
						GtkType           type);
const char       *gsk_lisp_namespace_get_nick  (GskLispNamespace *ns,
						GtkType           type);

/* --- implementation helpers --- */
void gsk_lisp_prefix_err_msg (char **msg, const char *fmt, ...);

/* --- debugging --- */
#ifdef GSK_DEBUG
GskLispCodec     *gsk_lisp_namespace_compile_debug(GskLispNamespace *ns,
						   GskLispNode      *node,
						   GtkType           type,
						   char            **err_msg,
						   const char       *file,
						   int               line);
#define gsk_lisp_namespace_compile(ns,node,type,err_msg)		\
	gsk_lisp_namespace_compile_debug(ns,node,type,err_msg,__FILE__,__LINE__)
#endif
 
#endif
